@@ -1,6 +1,6 @@
 /***********************************************************************
  *  avra - Assembler for the Atmel AVR microcontroller series
- *  Copyright (C) 1998-2000 Jon Anders Haugum
+ *  Copyright (C) 1998-2001 Jon Anders Haugum
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -26,12 +26,13 @@
 #include <stdio.h>
 
 #define VERSION 0
-#define REVISION 5
-#define DATESTRING "4 January 2000"
+#define REVISION 7
+#define DATESTRING "25 September 2001"
 
 #define IS_HOR_SPACE(x) ((x == ' ') || (x == 9))
 #define IS_LABEL(x) (isalnum(x) || (x == '_'))
 #define IS_END(x) ((x == ';') || (x == 10) || (x == 13)|| (x == '\0'))
+#define IS_ENDLINE(x) ( (x == 10) || (x == 13)|| (x == '\0'))
 
 #define LINEBUFFER_LENGTH 256
 
@@ -40,9 +41,12 @@
 /* Option enumeration */
 enum
 	{
-	ARG_DEFINE = 0,       /* --define                */
+	ARG_DEFINE = 0,       /* --define, -D            */
+	ARG_INCLUDEDIR,       /* --includedir, -I        */
 	ARG_LISTMAC,          /* --listmac               */
 	ARG_MAX_ERRORS,       /* --max_errors            */
+	ARG_COFF,             /* --coff                  */
+	ARG_DEVICES,          /* --devices               */
 	ARG_VER,              /* --version               */
 	ARG_HELP,             /* --help, -h              */
 	ARG_COUNT
@@ -74,7 +78,10 @@ enum
 	TERM_END = 0,
 	TERM_SPACE,
 	TERM_COMMA,
-	TERM_EQUAL
+	TERM_EQUAL,
+	TERM_DASH,
+	TERM_DOUBLEQUOTE,
+	TERM_COLON
 	};
 
 /* Structures */
@@ -117,6 +124,8 @@ struct prog_info
 	struct macro_call *first_macro_call;
 	struct macro_call *last_macro_call;
 	int conditional_depth;
+	/* coff additions */
+	FILE *coff_file;
 	};
 
 struct file_info
@@ -135,6 +144,7 @@ struct hex_file_info
 	FILE *fp;
 	int count;
 	int linestart_addr;
+	int segment;
 	unsigned char hex_line[16];
 	};
 
@@ -219,6 +229,8 @@ int get_mnemonic_type(char *mnemonic);
 int get_register(struct prog_info *pi, char *data);
 int get_bitnum(struct prog_info *pi, char *data, int *ret);
 int get_indirect(struct prog_info *pi, char *operand);
+int is_supported(struct prog_info *pi, char *name);
+int count_supported_instructions(int flags);
 
 /* directiv.c */
 int parse_directive(struct prog_info *pi, int pass);
@@ -228,6 +240,7 @@ int parse_db(struct prog_info *pi, char *next, int pass);
 void write_db(struct prog_info *pi, char byte, char *prev, int count, int pass);
 int spool_conditional(struct prog_info *pi, int only_endif);
 int check_conditional(struct prog_info *pi, char *buff, int *current_depth, int *do_next, int only_endif);
+int test_include(const char *filename);
 
 /* macro.c */
 int read_macro(struct prog_info *pi, char *name, int pass);
@@ -257,15 +270,14 @@ char *nocase_strstr(char *s, char *t);
 int atox(char *s);
 int atoi_n(char *s, int n);
 int atox_n(char *s, int n);
+char *my_strlwr(char *in);
+char *my_strupr(char *in);
 
-
-
-
-
-
-
-
-
-
-
-
+/* coff.c */
+FILE *open_coff_file(struct prog_info *pi, char *filename);
+void write_coff_file(struct prog_info *pi);
+void write_coff_eeprom( struct prog_info *pi, int address, unsigned char data);
+void write_coff_program( struct prog_info *pi, int address, unsigned int data);
+void close_coff_file(struct prog_info *pi, FILE *fp);
+int parse_stabs( struct prog_info *pi, char *p, int pass );
+int parse_stabn( struct prog_info *pi, char *p, int pass );
