@@ -1,6 +1,6 @@
 /***********************************************************************
  *  avra - Assembler for the Atmel AVR microcontroller series
- *  Copyright (C) 1998-2000 Jon Anders Haugum
+ *  Copyright (C) 1998-2001 Jon Anders Haugum
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -37,6 +37,7 @@ enum
 	{
 	DIRECTIVE_BYTE = 0,
 	DIRECTIVE_CSEG,
+	DIRECTIVE_CSEGSIZE,
 	DIRECTIVE_DB,
 	DIRECTIVE_DEF,
 	DIRECTIVE_DEVICE,
@@ -72,6 +73,7 @@ char *directive_list[] =
 	{
 	"BYTE",
 	"CSEG",
+	"CSEGSIZE",
 	"DB",
 	"DEF",
 	"DEVICE",
@@ -110,6 +112,7 @@ int parse_directive(struct prog_info *pi, int pass)
 	struct file_info *fi_bak;
 	struct label *label;
 	struct def *def;
+	struct data_list *includedir;
 
 	next = get_next_token(pi->fi->scratch, TERM_SPACE);
 	for(i = 0; pi->fi->scratch[i] != '\0'; i++)
@@ -145,6 +148,8 @@ int parse_directive(struct prog_info *pi, int pass)
 		case DIRECTIVE_CSEG:
 			pi->segment = SEGMENT_CODE;
 			break;
+		case DIRECTIVE_CSEGSIZE:
+			break;
 		case DIRECTIVE_DB:
 			return(parse_db(pi, next, pass));
 			break;
@@ -199,7 +204,7 @@ int parse_directive(struct prog_info *pi, int pass)
 				return(True);
 				}
 			get_next_token(next, TERM_END);
-			pi->device = get_device(next);
+			pi->device = get_device(pi,next);
 			if(!pi->device)
 				print_msg(pi, MSGTYPE_ERROR, "Unknown device: %s", next);
 			break;
@@ -220,7 +225,7 @@ int parse_directive(struct prog_info *pi, int pass)
 				  if(!get_expr(pi, next, &i))
 				    return(False);
 				  if((i < -32768) || (i > 65535))
-				    print_msg(pi, MSGTYPE_ERROR, "Value %d is out of range (-32768 <= k <= 65535)", i);
+				    print_msg(pi, MSGTYPE_WARNING, "Value %d is out of range (-32768 <= k <= 65535). Will be masked", i);
 				  }
 				if(pi->segment == SEGMENT_EEPROM)
 					{
@@ -322,9 +327,37 @@ int parse_directive(struct prog_info *pi, int pass)
 				fprintf(pi->list_file, "          %s", pi->list_line);
 				pi->list_line = NULL;
 				}
-			fi_bak = pi->fi;
-			ok = parse_file(pi, next, pass);
-			pi->fi = fi_bak;
+			// Test if include is in local directory
+			ok = test_include(next);
+			data = NULL;
+			if(!ok)
+				for(includedir = GET_ARG(pi->args, ARG_INCLUDEDIR); includedir && !ok; includedir = includedir->next)
+				{
+					i = strlen(includedir->data);
+					if(data)
+						free(data);
+					data = malloc(i + strlen(next) + 2);
+					if(!data)
+					{
+						print_msg(pi, MSGTYPE_OUT_OF_MEM, NULL);
+						return(False);
+					}
+					strcpy(data, includedir->data);
+					if((data[i - 1] != '\\') && (data[i - 1] != '/'))
+						data[i++] = '/';
+					strcpy(&data[i], next);
+					ok = test_include(data);
+				}
+			if(ok)
+			{
+				fi_bak = pi->fi;
+				ok = parse_file(pi, data ? data : next, pass);
+				pi->fi = fi_bak;
+			}
+			else
+				perror(next);
+			if(data)
+				free(data);
 			break;
 		case DIRECTIVE_LIST:
 			if(pass == PASS_2)
@@ -627,7 +660,7 @@ int parse_db(struct prog_info *pi, char *next, int pass)
 			  if(!get_expr(pi, next, &i))
 			    return(False);
 			  if((i < -128) || (i > 255))
-			    print_msg(pi, MSGTYPE_ERROR, "Value %d is out of range (-128 <= k <= 255)", i);
+			    print_msg(pi, MSGTYPE_WARNING, "Value %d is out of range (-128 <= k <= 255). Will be masked", i);
 			  }
 			count++;
 			write_db(pi, (char)i, &prev, count, pass);
@@ -639,7 +672,10 @@ int parse_db(struct prog_info *pi, char *next, int pass)
 		if((count % 2) == 1)
 			{
 			if(pass == PASS_2)
+			{
 				write_prog_word(pi, pi->cseg_addr, prev & 0xff);
+				print_msg(pi, MSGTYPE_WARNING, "A .DB segment with an odd number of bytes is detected. A zero byte is added.");
+			}
 			pi->cseg_addr++;
 			if(pass == PASS_1)
 				pi->cseg_count++;
@@ -772,3 +808,16 @@ int check_conditional(struct prog_info *pi, char *buff, int *current_depth, int
 	*do_next = True;
 	return(True);
 	}
+
+int test_include(const char *filename)
+{
+  FILE *fp;
+  fp = fopen(filename, "r");
+  if(fp)
+  {
+    fclose(fp);
+    return(True);
+  }
+  else
+    return(False);
+}
