@@ -1,6 +1,6 @@
 /***********************************************************************
  *  avra - Assembler for the Atmel AVR microcontroller series
- *  Copyright (C) 1998-2000 Jon Anders Haugum
+ *  Copyright (C) 1998-2001 Jon Anders Haugum
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,6 +30,7 @@
 
 #include "misc.h"
 #include "avra.h"
+#include "args.h"
 
 
 int open_out_files(struct prog_info *pi, char *filename)
@@ -37,6 +38,7 @@ int open_out_files(struct prog_info *pi, char *filename)
 	int length;
 	char *buff;
 	time_t tp;
+	int ok; /* flag for coff results */
 
 	length = strlen(filename);
 	buff = malloc(length + 9);
@@ -63,8 +65,23 @@ int open_out_files(struct prog_info *pi, char *filename)
 			strcpy(&buff[length], ".eep.hex");
 			pi->eep_hfi = open_hex_file(buff);
 			}
+
+		if(GET_ARG(pi->args, ARG_COFF)){
+			strcpy(&buff[length], ".cof");
+			pi->coff_file = open_coff_file(pi, buff);
+			if( pi->coff_file )
+				ok = True;
+			else
+				ok = False;
+		}
+		else{
+			pi->coff_file = 0;
+			ok = True;
+		}
+
 		free(buff);
-		if(pi->list_file && pi->obj_file && (!pi->cseg_count || pi->hfi) && (!pi->eseg_count || pi->eep_hfi))
+
+		if(pi->list_file && pi->obj_file && (!pi->cseg_count || pi->hfi) && (!pi->eseg_count || pi->eep_hfi) && ok)
 			{
 			if(time(&tp) != -1)
 				fprintf(pi->list_file, "\navra   ver. %d.%d  %s %s\n\n", VERSION, REVISION, filename, ctime(&tp));
@@ -98,6 +115,7 @@ void close_out_files(struct prog_info *pi)
 		fclose(pi->list_file);
 		}
 	if(pi->obj_file) close_obj_file(pi, pi->obj_file);
+	if(pi->coff_file) close_coff_file(pi, pi->coff_file);
 	}
 
 
@@ -108,6 +126,7 @@ struct hex_file_info *open_hex_file(char *filename)
 	hfi = calloc(1, sizeof(struct hex_file_info));
 	if(hfi)
 		{
+		hfi->segment = -1;
 		hfi->fp = fopen(filename, "wb");
 		if(!hfi->fp)
 			{
@@ -121,22 +140,10 @@ struct hex_file_info *open_hex_file(char *filename)
 
 void close_hex_file(struct hex_file_info *hfi)
 	{
-	int i;
-	unsigned char checksum = 0;
-
 	if(hfi->fp)
 		{
 		if(hfi->count != 0)
-			{
-			fprintf(hfi->fp, ":%02X%04X00", hfi->count, hfi->linestart_addr);
-			checksum -= hfi->count + ((hfi->linestart_addr >> 8) & 0xff) + (hfi->linestart_addr & 0xff);
-			for(i = 0; i < hfi->count; i++)
-				{
-				fprintf(hfi->fp, "%02X", hfi->hex_line[i]);
-				checksum -= hfi->hex_line[i];
-				}
-			fprintf(hfi->fp, "%02X\x0d\x0a", checksum);
-			}
+		  do_hex_line(hfi);
 		fprintf(hfi->fp, ":00000001FF\x0d\x0a");
 		fclose(hfi->fp);
 		}
@@ -151,6 +158,9 @@ void write_ee_byte(struct prog_info *pi, int address, unsigned char data)
 	if(pi->eep_hfi->count == 0)
 		pi->eep_hfi->linestart_addr = address;
 	pi->eep_hfi->hex_line[pi->eep_hfi->count++] = data;
+
+	if(pi->coff_file)
+	        write_coff_eeprom(pi, address, data);
 	}
 
 
@@ -158,12 +168,27 @@ void write_prog_word(struct prog_info *pi, int address, int data)
 	{
 	write_obj_record(pi, address, data);
 	address *= 2;
+	if(pi->hfi->segment != (address >> 16))
+	{
+	  if(pi->hfi->count != 0)
+	    do_hex_line(pi->hfi);
+	  pi->hfi->segment = address >> 16;
+	  if(pi->hfi->segment >= 16) // Use 04 record for addresses above 1 meg since 02 can support max 1 meg
+	    fprintf(pi->hfi->fp, ":02000004%04X%02X\x0d\x0a", pi->hfi->segment & 0xffff,
+	            (0 - 2 - 4 - ((pi->hfi->segment >> 8) & 0xff) - (pi->hfi->segment & 0xff)) & 0xff);
+	  else // Use 02 record for addresses below 1 meg since more programmers know about the 02 instead of the 04
+	    fprintf(pi->hfi->fp, ":02000002%04X%02X\x0d\x0a", (pi->hfi->segment << 12) & 0xffff,
+	            (0 - 2 - 2 - ((pi->hfi->segment << 4) & 0xf0)) & 0xff);
+	}
 	if((pi->hfi->count == 16) || ((address != (pi->hfi->linestart_addr + pi->hfi->count)) && (pi->hfi->count != 0)))
 		do_hex_line(pi->hfi);
 	if(pi->hfi->count == 0)
 		pi->hfi->linestart_addr = address;
 	pi->hfi->hex_line[pi->hfi->count++] = data & 0xff;
 	pi->hfi->hex_line[pi->hfi->count++] = (data >> 8) & 0xff;
+
+	if(pi->coff_file != 0)
+	        write_coff_program(pi, address, data);
 	}
 
 
@@ -172,7 +197,7 @@ void do_hex_line(struct hex_file_info *hfi)
 	int i;
 	unsigned char checksum = 0;
 
-	fprintf(hfi->fp, ":%02X%04X00", hfi->count, hfi->linestart_addr);
+	fprintf(hfi->fp, ":%02X%04X00", hfi->count, hfi->linestart_addr & 0xffff);
 	checksum -= hfi->count + ((hfi->linestart_addr >> 8) & 0xff) + (hfi->linestart_addr & 0xff);
 	for(i = 0; i < hfi->count; i++)
 		{
