@@ -12,7 +12,7 @@ void write_map_file(struct prog_info *pi)
 	char File[200],*P;
 
 	strcpy(File,(char *)pi->args->first_data->data);
-	P = strstr(File,".");
+	P = strrchr(File,'.');
 	if( P ) *P = 0;
 	strcat(File,".map");
 	fp = fopen(File,"w");
@@ -42,4 +42,3 @@ char *Space(char *n) {
 	if( i < 8 ) return "\t\t";
 	return "\t";
 }
-
