@@ -1,6 +1,6 @@
 /***********************************************************************
  *  avra - Assembler for the Atmel AVR microcontroller series
- *  Copyright (C) 1998-2000 Jon Anders Haugum
+ *  Copyright (C) 1998-2001 Jon Anders Haugum
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -32,6 +32,8 @@
 #include "avra.h"
 #include "device.h"
 
+#define MAX_MNEMONIC_LEN	8	// Maximum mnemonic length
+
 enum
 	{
 	MNEMONIC_NOP = 0,  //          0000 0000 0000 0000
@@ -51,12 +53,17 @@ enum
 	MNEMONIC_CLT,      //          1001 0100 1110 1000
 	MNEMONIC_SEH,      //          1001 0100 0101 1000
 	MNEMONIC_CLH,      //          1001 0100 1101 1000
-	MNEMONIC_SLEEP,    //          1001 0101 100X 1000
-	MNEMONIC_WDR,      //          1001 0101 101X 1000
-	MNEMONIC_IJMP,     //          1001 0100 XXXX 1001
-	MNEMONIC_ICALL,    //          1001 0101 XXXX 1001
-	MNEMONIC_RET,      //          1001 0101 0XX0 1000
-	MNEMONIC_RETI,     //          1001 0101 0XX1 1000
+	MNEMONIC_SLEEP,    //          1001 0101 1000 1000
+	MNEMONIC_WDR,      //          1001 0101 1010 1000
+	MNEMONIC_IJMP,     //          1001 0100 0000 1001
+	MNEMONIC_EIJMP,    //          1001 0100 0001 1001
+	MNEMONIC_ICALL,    //          1001 0101 0000 1001
+	MNEMONIC_EICALL,   //          1001 0101 0001 1001
+	MNEMONIC_RET,      //          1001 0101 0000 1000
+	MNEMONIC_RETI,     //          1001 0101 0001 1000
+	MNEMONIC_SPM,      //          1001 0101 1110 1000
+	MNEMONIC_ESPM,     //          1001 0101 1111 1000
+	MNEMONIC_BREAK,    //          1001 0101 1001 1000
 	MNEMONIC_LPM,      //          1001 0101 1100 1000
 	MNEMONIC_ELPM,     //          1001 0101 1101 1000
 	MNEMONIC_BSET,     // s        1001 0100 0sss 1000
@@ -112,6 +119,12 @@ enum
 	MNEMONIC_CPSE,     // Rd, Rr   0001 00rd dddd rrrr
 	MNEMONIC_MOV,      // Rd, Rr   0010 11rd dddd rrrr
 	MNEMONIC_MUL,      // Rd, Rr   1001 11rd dddd rrrr
+	MNEMONIC_MOVW,     // Rd, Rr   0000 0001 dddd rrrr
+	MNEMONIC_MULS,     // Rd, Rr   0000 0010 dddd rrrr
+	MNEMONIC_MULSU,    // Rd, Rr   0000 0011 0ddd 0rrr
+	MNEMONIC_FMUL,     // Rd, Rr   0000 0011 0ddd 1rrr
+	MNEMONIC_FMULS,    // Rd, Rr   0000 0011 1ddd 0rrr
+	MNEMONIC_FMULSU,   // Rd, Rr   0000 0011 1ddd 1rrr
 	MNEMONIC_ADIW,     // Rd, K    1001 0110 KKdd KKKK
 	MNEMONIC_SBIW,     // Rd, K    1001 0111 KKdd KKKK
 	MNEMONIC_SUBI,     // Rd, K    0101 KKKK dddd KKKK
@@ -122,9 +135,9 @@ enum
 	MNEMONIC_CPI,      // Rd, K    0011 KKKK dddd KKKK
 	MNEMONIC_LDI,      // Rd, K    1110 KKKK dddd KKKK
 	MNEMONIC_CBR,      // Rd, K    0111 KKKK dddd KKKK ~K
-	MNEMONIC_SBRC,     // Rr, b    1111 110r rrrr Xbbb
-	MNEMONIC_SBRS,     // Rr, b    1111 111r rrrr Xbbb
-	MNEMONIC_BST,      // Rr, b    1111 101d dddd Xbbb
+	MNEMONIC_SBRC,     // Rr, b    1111 110r rrrr 0bbb
+	MNEMONIC_SBRS,     // Rr, b    1111 111r rrrr 0bbb
+	MNEMONIC_BST,      // Rr, b    1111 101d dddd 0bbb
 	MNEMONIC_BLD,      // Rd, b    1111 100d dddd 0bbb
 	MNEMONIC_IN,       // Rd, P    1011 0PPd dddd PPPP
 	MNEMONIC_OUT,      // P, Rr    1011 1PPr rrrr PPPP
@@ -139,6 +152,10 @@ enum
 	MNEMONIC_LDD,      // Rd, _+q  dummy
 	MNEMONIC_STD,      // _+q, Rr  dummy
 	MNEMONIC_COUNT,
+	MNEMONIC_LPM_Z,    // Rd, Z    1001 000d dddd 0100
+	MNEMONIC_LPM_ZP,   // Rd, Z+   1001 000d dddd 0101
+	MNEMONIC_ELPM_Z,   // Rd, Z    1001 000d dddd 0110
+	MNEMONIC_ELPM_ZP,  // Rd, Z+   1001 000d dddd 0111
 	MNEMONIC_LD_X,     // Rd, X    1001 000d dddd 1100
 	MNEMONIC_LD_XP,    // Rd, X+   1001 000d dddd 1101
 	MNEMONIC_LD_MX,    // Rd, -X   1001 000d dddd 1110
@@ -161,143 +178,162 @@ enum
 	MNEMONIC_LDD_Z,    // Rd, Z+q  10q0 qq0d dddd 0qqq
 	MNEMONIC_STD_Y,    // Y+q, Rr  10q0 qq1r rrrr 1qqq
 	MNEMONIC_STD_Z,    // Z+q, Rr  10q0 qq1r rrrr 0qqq
+	MNEMONIC_END
 	};
 
 struct instruction
 	{
 	char *mnemonic;
 	int opcode;
+        int flag;	/* Device flags meaning the instruction is not
+                           supported */
 	};
 
 struct instruction instruction_list[] =
 	{
-	{"nop",   0x0000},
-	{"sec",   0x9408},
-	{"clc",   0x9488},
-	{"sen",   0x9428},
-	{"cln",   0x94a8},
-	{"sez",   0x9418},
-	{"clz",   0x9498},
-	{"sei",   0x9478},
-	{"cli",   0x94f8},
-	{"ses",   0x9448},
-	{"cls",   0x94c8},
-	{"sev",   0x9438},
-	{"clv",   0x94b8},
-	{"set",   0x9468},
-	{"clt",   0x94e8},
-	{"seh",   0x9458},
-	{"clh",   0x94d8},
-	{"sleep", 0x9588},
-	{"wdr",   0x95a8},
-	{"ijmp",  0x9409},
-	{"icall", 0x9509},
-	{"ret",   0x9508},
-	{"reti",  0x9518},
-	{"lpm",   0x95c8},
-	{"elpm",  0x95d8},
-	{"bset",  0x9408},
-	{"bclr",  0x9488},
-	{"ser",   0xef0f},
-	{"com",   0x9400},
-	{"neg",   0x9401},
-	{"inc",   0x9403},
-	{"dec",   0x940a},
-	{"lsr",   0x9406},
-	{"ror",   0x9407},
-	{"asr",   0x9405},
-	{"swap",  0x9402},
-	{"push",  0x920f},
-	{"pop",   0x900f},
-	{"tst",   0x2000},
-	{"clr",   0x2400},
-	{"lsl",   0x0c00},
-	{"rol",   0x1c00},
-	{"breq",  0xf001},
-	{"brne",  0xf401},
-	{"brcs",  0xf000},
-	{"brcc",  0xf400},
-	{"brsh",  0xf400},
-	{"brlo",  0xf000},
-	{"brmi",  0xf002},
-	{"brpl",  0xf402},
-	{"brge",  0xf404},
-	{"brlt",  0xf004},
-	{"brhs",  0xf005},
-	{"brhc",  0xf405},
-	{"brts",  0xf006},
-	{"brtc",  0xf406},
-	{"brvs",  0xf003},
-	{"brvc",  0xf403},
-	{"brie",  0xf007},
-	{"brid",  0xf407},
-	{"rjmp",  0xc000},
-	{"rcall", 0xd000},
-	{"jmp",   0x940c},
-	{"call",  0x940e},
-	{"brbs",  0xf000},
-	{"brbc",  0xf400},
-	{"add",   0x0c00},
-	{"adc",   0x1c00},
-	{"sub",   0x1800},
-	{"sbc",   0x0800},
-	{"and",   0x2000},
-	{"or",    0x2800},
-	{"eor",   0x2400},
-	{"cp",    0x1400},
-	{"cpc",   0x0400},
-	{"cpse",  0x1000},
-	{"mov",   0x2c00},
-	{"mul",   0x9c00},
-	{"adiw",  0x9600},
-	{"sbiw",  0x9700},
-	{"subi",  0x5000},
-	{"sbci",  0x4000},
-	{"andi",  0x7000},
-	{"ori",   0x6000},
-	{"sbr",   0x6000},
-	{"cpi",   0x3000},
-	{"ldi",   0xe000},
-	{"cbr",   0x7000},
-	{"sbrc",  0xfc00},
-	{"sbrs",  0xfe00},
-	{"bst",   0xfa00},
-	{"bld",   0xf800},
-	{"in",    0xb000},
-	{"out",   0xb800},
-	{"sbic",  0x9900},
-	{"sbis",  0x9b00},
-	{"sbi",   0x9a00},
-	{"cbi",   0x9800},
-	{"lds",   0x9000},
-	{"sts",   0x9200},
-	{"ld",    0},
-	{"st",    0},
-	{"ldd",   0},
-	{"std",   0},
-	{"count", 0},
-	{"ld",    0x900c},
-	{"ld",    0x900d},
-	{"ld",    0x900e},
-	{"ld",    0x8008},
-	{"ld",    0x9009},
-	{"ld",    0x900a},
-	{"ld",    0x8000},
-	{"ld",    0x9001},
-	{"ld",    0x9002},
-	{"st",    0x920c},
-	{"st",    0x920d},
-	{"st",    0x920e},
-	{"st",    0x8208},
-	{"st",    0x9209},
-	{"st",    0x920a},
-	{"st",    0x8200},
-	{"st",    0x9201},
-	{"st",    0x9202},
-	{"ldd",   0x8008},
-	{"ldd",   0x8000},
-	{"std",   0x8208},
-	{"std",   0x8200}
+	{"nop",   0x0000,          0},
+	{"sec",   0x9408,          0},
+	{"clc",   0x9488,          0},
+	{"sen",   0x9428,          0},
+	{"cln",   0x94a8,          0},
+	{"sez",   0x9418,          0},
+	{"clz",   0x9498,          0},
+	{"sei",   0x9478,          0},
+	{"cli",   0x94f8,          0},
+	{"ses",   0x9448,          0},
+	{"cls",   0x94c8,          0},
+	{"sev",   0x9438,          0},
+	{"clv",   0x94b8,          0},
+	{"set",   0x9468,          0},
+	{"clt",   0x94e8,          0},
+	{"seh",   0x9458,          0},
+	{"clh",   0x94d8,          0},
+	{"sleep", 0x9588,          0},
+	{"wdr",   0x95a8,          0},
+	{"ijmp",  0x9409,  DF_TINY1X},
+	{"eijmp", 0x9419, DF_NO_EIJMP},
+	{"icall", 0x9509,  DF_TINY1X},
+	{"eicall",0x9519, DF_NO_EICALL},
+	{"ret",   0x9508,          0},
+	{"reti",  0x9518,          0},
+	{"spm",   0x95e8, DF_NO_SPM},
+	{"espm",  0x95f8, DF_NO_ESPM},
+	{"break", 0x9598, DF_NO_BREAK},
+	{"lpm",   0x95c8, DF_NO_LPM},
+	{"elpm",  0x95d8, DF_NO_ELPM},
+	{"bset",  0x9408,          0},
+	{"bclr",  0x9488,          0},
+	{"ser",   0xef0f,          0},
+	{"com",   0x9400,          0},
+	{"neg",   0x9401,          0},
+	{"inc",   0x9403,          0},
+	{"dec",   0x940a,          0},
+	{"lsr",   0x9406,          0},
+	{"ror",   0x9407,          0},
+	{"asr",   0x9405,          0},
+	{"swap",  0x9402,          0},
+	{"push",  0x920f,  DF_TINY1X},
+	{"pop",   0x900f,  DF_TINY1X},
+	{"tst",   0x2000,          0},
+	{"clr",   0x2400,          0},
+	{"lsl",   0x0c00,          0},
+	{"rol",   0x1c00,          0},
+	{"breq",  0xf001,          0},
+	{"brne",  0xf401,          0},
+	{"brcs",  0xf000,          0},
+	{"brcc",  0xf400,          0},
+	{"brsh",  0xf400,          0},
+	{"brlo",  0xf000,          0},
+	{"brmi",  0xf002,          0},
+	{"brpl",  0xf402,          0},
+	{"brge",  0xf404,          0},
+	{"brlt",  0xf004,          0},
+	{"brhs",  0xf005,          0},
+	{"brhc",  0xf405,          0},
+	{"brts",  0xf006,          0},
+	{"brtc",  0xf406,          0},
+	{"brvs",  0xf003,          0},
+	{"brvc",  0xf403,          0},
+	{"brie",  0xf007,          0},
+	{"brid",  0xf407,          0},
+	{"rjmp",  0xc000,          0},
+	{"rcall", 0xd000,          0},
+	{"jmp",   0x940c,  DF_NO_JMP},
+	{"call",  0x940e,  DF_NO_JMP},
+	{"brbs",  0xf000,          0},
+	{"brbc",  0xf400,          0},
+	{"add",   0x0c00,          0},
+	{"adc",   0x1c00,          0},
+	{"sub",   0x1800,          0},
+	{"sbc",   0x0800,          0},
+	{"and",   0x2000,          0},
+	{"or",    0x2800,          0},
+	{"eor",   0x2400,          0},
+	{"cp",    0x1400,          0},
+	{"cpc",   0x0400,          0},
+	{"cpse",  0x1000,          0},
+	{"mov",   0x2c00,          0},
+	{"mul",   0x9c00, DF_NO_MUL},
+	{"movw",  0x0100, DF_NO_MOVW},
+	{"muls",  0x0200, DF_NO_MUL},
+	{"mulsu", 0x0300, DF_NO_MUL},
+	{"fmul",  0x0308, DF_NO_MUL},
+	{"fmuls", 0x0380, DF_NO_MUL},
+	{"fmulsu",0x0388, DF_NO_MUL},
+	{"adiw",  0x9600,  DF_TINY1X},
+	{"sbiw",  0x9700,  DF_TINY1X},
+	{"subi",  0x5000,          0},
+	{"sbci",  0x4000,          0},
+	{"andi",  0x7000,          0},
+	{"ori",   0x6000,          0},
+	{"sbr",   0x6000,          0},
+	{"cpi",   0x3000,          0},
+	{"ldi",   0xe000,          0},
+	{"cbr",   0x7000,          0},
+	{"sbrc",  0xfc00,          0},
+	{"sbrs",  0xfe00,          0},
+	{"bst",   0xfa00,          0},
+	{"bld",   0xf800,          0},
+	{"in",    0xb000,          0},
+	{"out",   0xb800,          0},
+	{"sbic",  0x9900,          0},
+	{"sbis",  0x9b00,          0},
+	{"sbi",   0x9a00,          0},
+	{"cbi",   0x9800,          0},
+	{"lds",   0x9000,  DF_TINY1X},
+	{"sts",   0x9200,  DF_TINY1X},
+	{"ld",    0,          0},
+	{"st",    0,          0},
+	{"ldd",   0,  DF_TINY1X},
+	{"std",   0,  DF_TINY1X},
+	{"count", 0,          0},
+	{"lpm",   0x9004, DF_NO_LPM|DF_NO_LPM_X},
+	{"lpm",   0x9005, DF_NO_LPM|DF_NO_LPM_X},
+	{"elpm",  0x9006, DF_NO_ELPM|DF_NO_ELPM_X},
+	{"elpm",  0x9007, DF_NO_ELPM|DF_NO_ELPM_X},
+	{"ld",    0x900c, DF_NO_XREG},
+	{"ld",    0x900d, DF_NO_XREG},
+	{"ld",    0x900e, DF_NO_XREG},
+	{"ld",    0x8008, DF_NO_YREG},
+	{"ld",    0x9009, DF_NO_YREG},
+	{"ld",    0x900a, DF_NO_YREG},
+	{"ld",    0x8000,          0},
+	{"ld",    0x9001, DF_TINY1X},
+	{"ld",    0x9002, DF_TINY1X},
+	{"st",    0x920c, DF_NO_XREG},
+	{"st",    0x920d, DF_NO_XREG},
+	{"st",    0x920e, DF_NO_XREG},
+	{"st",    0x8208, DF_NO_YREG},
+	{"st",    0x9209, DF_NO_YREG},
+	{"st",    0x920a, DF_NO_YREG},
+	{"st",    0x8200,          0},
+	{"st",    0x9201, DF_TINY1X},
+	{"st",    0x9202, DF_TINY1X},
+	{"ldd",   0x8008, DF_TINY1X},
+	{"ldd",   0x8000, DF_TINY1X},
+	{"std",   0x8208, DF_TINY1X},
+	{"std",   0x8200, DF_TINY1X},
+	{"end", 0, 0}
 	};
 
 
@@ -306,11 +342,10 @@ int parse_mnemonic(struct prog_info *pi, int pass)
 	int mnemonic, i, opcode, opcode2, instruction_long = False;
 	char *operand1, *operand2;
 	struct macro *macro;
+	char temp[MAX_MNEMONIC_LEN+1];
 
 	operand1 = get_next_token(pi->fi->scratch, TERM_SPACE);
-	for(i = 0; pi->fi->scratch[i] != '\0'; i++)
-		pi->fi->scratch[i] = tolower(pi->fi->scratch[i]);
-	mnemonic = get_mnemonic_type(pi->fi->scratch);
+	mnemonic = get_mnemonic_type(my_strlwr(pi->fi->scratch));
 	if(mnemonic == -1)
 		{
 		macro = get_macro(pi, pi->fi->scratch);
@@ -326,11 +361,50 @@ int parse_mnemonic(struct prog_info *pi, int pass)
 		}
 	if(pass == PASS_2)
 		{
-		if(mnemonic <= MNEMONIC_ELPM)
+		if(mnemonic <= MNEMONIC_BREAK)
 			{
+			if(operand1)
+				print_msg(pi, MSGTYPE_WARNING, "Garbage after instruction %s: %s", instruction_list[mnemonic].mnemonic, operand1);
 			// No operand
 			opcode = 0;
 			}
+		else if(mnemonic <= MNEMONIC_ELPM)
+		{
+			if(operand1)
+			{
+				operand2 = get_next_token(operand1, TERM_COMMA);
+				if(!operand2)
+				{
+					print_msg(pi, MSGTYPE_ERROR, "%s needs a second operand", instruction_list[mnemonic].mnemonic);
+					return(True);
+				}
+				get_next_token(operand2, TERM_END);
+				i = get_register(pi, operand1);
+				opcode = i << 4;
+				i = get_indirect(pi, operand2);
+				if(i == 6) // Means Z
+				{
+					if(mnemonic == MNEMONIC_LPM)
+						mnemonic = MNEMONIC_LPM_Z;
+					else if(mnemonic == MNEMONIC_ELPM)
+						mnemonic = MNEMONIC_ELPM_Z;
+				}
+				else if(i == 7) // Means Z+
+				{
+					if(mnemonic == MNEMONIC_LPM)
+						mnemonic = MNEMONIC_LPM_ZP;
+					else if(mnemonic == MNEMONIC_ELPM)
+						mnemonic = MNEMONIC_ELPM_ZP;
+				}
+				else
+				{
+					print_msg(pi, MSGTYPE_ERROR, "Unsupported operand: %s", operand2);
+					return(True);
+				}
+			}
+			else
+				opcode = 0;
+		}
 		else
 			{
 			if(!operand1)
@@ -408,13 +482,44 @@ int parse_mnemonic(struct prog_info *pi, int pass)
 				}
 			else if(mnemonic <= MNEMONIC_MUL)
 				{
-				if((mnemonic == MNEMONIC_MUL) && (pi->device->flag & DF_NO_MUL))
-					print_msg(pi, MSGTYPE_ERROR, "MUL instruction is not supported on %s", pi->device->name);
 				i = get_register(pi, operand1);
 				opcode = i << 4;
 				i = get_register(pi, operand2);
 				opcode |= ((i & 0x10) << 5) | (i & 0x0f);
 				}
+			else if(mnemonic <= MNEMONIC_MOVW)
+			{
+				i = get_register(pi, operand1);
+				if((i % 2) == 1)
+					print_msg(pi, MSGTYPE_ERROR, "%s must use a even numbered register for Rd", instruction_list[mnemonic].mnemonic);
+				opcode = (i / 2) << 4;
+				i = get_register(pi, operand2);
+				if((i % 2) == 1)
+					print_msg(pi, MSGTYPE_ERROR, "%s must use a even numbered register for Rr", instruction_list[mnemonic].mnemonic);
+				opcode |= i / 2;
+			}
+			else if(mnemonic <= MNEMONIC_MULS)
+			{
+				i = get_register(pi, operand1);
+				if(i < 16)
+					print_msg(pi, MSGTYPE_ERROR, "%s can only use a high register (r16 - r31)", instruction_list[mnemonic].mnemonic);
+				opcode = (i & 0x0f) << 4;
+				i = get_register(pi, operand2);
+				if(i < 16)
+					print_msg(pi, MSGTYPE_ERROR, "%s can only use a high register (r16 - r31)", instruction_list[mnemonic].mnemonic);
+				opcode |= (i & 0x0f);
+			}
+			else if(mnemonic <= MNEMONIC_FMULSU)
+			{
+				i = get_register(pi, operand1);
+				if((i < 16) || (i >= 24))
+					print_msg(pi, MSGTYPE_ERROR, "%s can only use registers (r16 - r23)", instruction_list[mnemonic].mnemonic);
+				opcode = (i & 0x07) << 4;
+				i = get_register(pi, operand2);
+				if((i < 16) || (i >= 24))
+					print_msg(pi, MSGTYPE_ERROR, "%s can only use registers (r16 - r23)", instruction_list[mnemonic].mnemonic);
+				opcode |= (i & 0x07);
+			}
 			else if(mnemonic <= MNEMONIC_SBIW)
 				{
 				i = get_register(pi, operand1);
@@ -436,7 +541,7 @@ int parse_mnemonic(struct prog_info *pi, int pass)
 				if(!get_expr(pi, operand2, &i))
 					return(False);
 				if((i < -128) || (i > 255))
-					print_msg(pi, MSGTYPE_ERROR, "Constant out of range (-128 <= k <= 255)");
+					print_msg(pi, MSGTYPE_WARNING, "Constant out of range (-128 <= k <= 255). Will be masked");
 				if(mnemonic == MNEMONIC_CBR)
 					i = ~i;
 				opcode |= ((i & 0xf0) << 4) | (i & 0x0f);
@@ -563,6 +668,12 @@ int parse_mnemonic(struct prog_info *pi, int pass)
 			else
 				print_msg(pi, MSGTYPE_ERROR, "Shit! Missing opcode check [%d]...", mnemonic);
 			}
+ 	        if (pi->device->flag & instruction_list[mnemonic].flag)
+		{
+			strncpy(temp, instruction_list[mnemonic].mnemonic, MAX_MNEMONIC_LEN);
+	   		print_msg(pi, MSGTYPE_ERROR, "%s instruction is not supported on %s",
+	   		          my_strupr(temp), pi->device->name);
+		}
 		opcode |= instruction_list[mnemonic].opcode;
 		if(pi->list_on && pi->list_line)
 			{
@@ -613,9 +724,15 @@ int get_mnemonic_type(char *mnemonic)
 
 int get_register(struct prog_info *pi, char *data)
 	{
+	char *second_reg;
 	int reg = 0;
 	struct def *def;
 
+	// Check for any occurence of r1:r0 pairs, and if so skip to second register
+	second_reg = strchr(data, ':');
+	if(second_reg != NULL)
+		data = second_reg + 1;
+
 	for(def = pi->first_def; def; def = def->next)
 		if(!nocase_strcmp(def->name, data))
 			{
@@ -635,7 +752,7 @@ int get_register(struct prog_info *pi, char *data)
 
 
 int get_bitnum(struct prog_info *pi, char *data, int *ret)
-	{
+{
 	if(!get_expr(pi, data, ret))
 		return(False);
 	if((*ret < 0) || (*ret > 7))
@@ -660,8 +777,12 @@ int get_indirect(struct prog_info *pi, char *operand)
 			switch(tolower(operand[i]))
 				{
 				case 'x':
+        	   	                if (pi->device->flag & DF_NO_XREG)
+                                           print_msg(pi, MSGTYPE_ERROR, "X register is not supported on %s", pi->device->name);
 					return(2);
 				case 'y':
+        	   	                if (pi->device->flag & DF_NO_YREG)
+                                           print_msg(pi, MSGTYPE_ERROR, "Y register is not supported on %s", pi->device->name);
 					return(5);
 				case 'z':
 					return(8);
@@ -670,6 +791,8 @@ int get_indirect(struct prog_info *pi, char *operand)
 					return(0);
 				}
 		case 'x':
+                        if (pi->device->flag & DF_NO_XREG)
+                           print_msg(pi, MSGTYPE_ERROR, "X register is not supported on %s", pi->device->name);
 			while(IS_HOR_SPACE(operand[i])) i++;
 			if(operand[i] == '+')
 				{
@@ -683,6 +806,8 @@ int get_indirect(struct prog_info *pi, char *operand)
 				print_msg(pi, MSGTYPE_ERROR, "Garbage after operand (%s)", operand);
 			return(0);
 		case 'y':
+                        if (pi->device->flag & DF_NO_YREG)
+                           print_msg(pi, MSGTYPE_ERROR, "Y register is not supported on %s", pi->device->name);
 			while(IS_HOR_SPACE(operand[i])) i++;
 			if(operand[i] == '+')
 				{
@@ -713,3 +838,29 @@ int get_indirect(struct prog_info *pi, char *operand)
 		}
 	return(0);
 	}
+
+/* Return 1 if instruction name is supported by the current device,
+   0 if unsupported, -1 if it is invalid */
+int is_supported(struct prog_info *pi, char *name) {
+   char temp[MAX_MNEMONIC_LEN+1];
+   int mnemonic;
+
+   strncpy(temp,name,MAX_MNEMONIC_LEN);
+   mnemonic = get_mnemonic_type(my_strlwr(temp));
+   if (mnemonic == -1) return -1;
+   if (pi->device->flag & instruction_list[mnemonic].flag) return 0;
+   return 1;
+   }
+
+int count_supported_instructions(int flags)
+{
+  int i = 0, count = 0;
+  while(i < MNEMONIC_END)
+  {
+    if((i < MNEMONIC_LD) || (i > MNEMONIC_COUNT))
+      if(!(flags & instruction_list[i].flag))
+        count++;
+    i++;
+  }
+  return(count);
+}
